###############################################################################
ConstructHashes <- function(dataFrame, idColumnName, hashColumnName,
                            hash = "md5", experimentString = "ExperimentName") {
  require("digest") # computes md5 hash
  require("Rmpfr")  # lets convert hash to numbers
  stopifnot(idColumnName %in% names(dataFrame))
  if (hashColumnName %in% names(dataFrame))
    warning(paste("Column", hashColumnName, "exists in dataFrame,",
                  "it will be overwritten!"))
  # start work by initializin column
  dataFrame[[hashColumnName]] <- ""
  # not vectorized, gotta loop :(
  for (i in seq_len(nrow(dataFrame))) {
    itemIdString <- paste0(experimentString, dataFrame[[idColumnName]][[i]])
    dataFrame[[hashColumnName]][[i]] <- digest::digest(
      itemIdString, algo = hash, serialize = FALSE, seed = 1)
  } # end for each row
  # convert hash string into number and scale them to unit simplex
  dataFrame[["hashNumeric"]] <- Rmpfr::mpfr(dataFrame[[hashColumnName]],
                                            base = 16)
  dataFrame[["hashScaled"]] <- as.numeric(dataFrame[["hashNumeric"]] /
                                            max(dataFrame[["hashNumeric"]]))
  dataFrame
} # end ConstructMd5Hashes
###############################################################################
PerformKSTest <- function(dataVector, quantilesToCompute) {
  currentWarning <- options()[["warn"]]
  options(warn = -1)
  exactTest <- ks.test(dataVector, "punif", alternative = "two.sided")
  dataQuantiles <- quantile(dataVector, names = FALSE,
                            probs = quantilesToCompute)
  results <- data.frame(hashValue        = dataQuantiles,
                        uniformValue     = quantilesToCompute,
                        probability      = quantilesToCompute,
                        stringsAsFactors = FALSE)
  # approxTest <- ks.test(results[["hashValue"]], results[["uniformValue"]],
  #                       alternative = "two.sided")
  result <- data.frame(testType = c("KS-Exact")#, "KS-Approximate"),
                       , testStatistic = c(exactTest[["statistic"]])#,
                                         #approxTest[["statistic"]]),
                       , testPValue = c(exactTest[["p.value"]])#
                                      #approxTest[["p.value"]]),
                       , stringsAsFactors = FALSE)
  # print(result)
  result <- gather(result, key, value, testStatistic, testPValue)
  options(warn = currentWarning)
  result
} # end PerformKSTest
###############################################################################
PerformChiSquaredTest <- function(dataVector, roundFactor = 0.05) {
  require("dplyr")
  stopifnot(is.null(dim(dataVector)) && is.atomic(dataVector))
  dataFrame <- data.frame(dataVector = dataVector, stringsAsFactors = FALSE) %>%
    mutate(
      hashQuantile = roundFactor * floor((dataVector + roundFactor) /
                                           roundFactor)
      , hashQuantile = ifelse(hashQuantile > 1, 1, hashQuantile)
    ) %>%
    group_by(hashQuantile) %>%
    summarize(observedCounts = n()) %>%
    mutate(
      expectedCounts = length(dataVector) / nrow(.)
      , squaredDifferences = (observedCounts - expectedCounts)^2
      , chiSquared = squaredDifferences / expectedCounts
    )
  chiSquaredStatistic <- sum(dataFrame[["chiSquared"]])
  result <- data.frame(
    testType        = "ChiSq-Exact"
    , testStatistic = chiSquaredStatistic
    , testPValue    = pchisq(chiSquaredStatistic, df = nrow(dataFrame),
                             lower.tail = FALSE)
    , stringsAsFactors = FALSE
  )
  # print(result)
  result <- gather(result, key, value, testStatistic, testPValue)
  result
} # end PerformChiSquaredTest()
###############################################################################
LoadDominicsData <- function(dataDirPath, fileExtension = "csv") {
  require("readr")
  require("dplyr")
  listOfFiles <- list.files(path = dataDirPath, pattern = fileExtension)
  allData <- vector("list", length(listOfFiles))
  # load data from files
  for (i in seq_along(listOfFiles)) {
  fileName <- file.path(dataDirPath, listOfFiles[[i]])
    fileType <- str_replace_all(fileName, str_c(dataDirPath, "/"), "") %>%
      str_replace_all(str_c(".", fileExtension), "") %>%
      str_replace_all("upc", "")
    data <- read_csv(fileName, col_types = "cccccc")
    names(data) <- tolower(names(data))
    stopifnot(sum(is.na(data[["upc"]])) == 0)
    result <- data %>%
      mutate(category = fileType) %>%
      select(category, com_code, upc, descrip)
    allData[[i]] <- result
  } # end for
  allUpcs <- do.call(rbind, allData)
  allUpcs
} # end LoadDominicsData()
###############################################################################
CleanDominicsData <- function(dataFrame) {
  stopifnot(all(names(dataFrame) %in% c("category", "com_code", "upc",
                                        "descrip")))
  # some upcs are in more than one com_code, we must fix that for now
  # later will try redefine upc as com_code + upc to keep all rows
  # likely won't matter much
  catCountByUpc <- dataFrame %>%
    group_by(upc) %>%
    summarize(n = n())
  cleanUpcs <- dataFrame %>%
    group_by(upc) %>%
    mutate(rownum = seq_len(n())) %>%
    ungroup() %>%
    inner_join(catCountByUpc, by = "upc") %>%
    filter(rownum == 1) %>%
    select(-one_of("rownum", "n")) %>%
    # mutate(upc = as.numeric(upc)) %>%
    assert(not_na, upc)
  stopifnot(nrow(cleanUpcs) == length(unique(cleanUpcs[["upc"]])))

  # all upcs must be 12 digits long, since none here are, we will prepend
  # zeros to them where needed. Likely this will also have to be explored
  # later
  result <- cleanUpcs %>%
    mutate(
      upcLength = nchar(upc),
      originalUpc = upc,
      upc = str_c(str_dup("0", 12 - upcLength), originalUpc)
    ) %>%
    select(-upcLength)
   result
} # CleanDominicsData()
###############################################################################
